// roughly from: 
//   http://davebollinger.com/works/fungi/applet_fungi01/Fungi01.pde.txt

FixedPointBuffer buf;
float cx, cy, sc;
int ncps = 5000;
int niters = 16;

float pa, pb, pc, pd;

int myv;
float mya;

float sint;
float cost;

int lastT;

void setup() {
  size(1024, 768, P3D);
  //size(1680,1050);
  buf = new FixedPointBuffer(width, height);
  cx = width / 2.0;
  cy = height / 2.0;
  sc = width / (2*PI);

  myv = 255 << 16;
  mya = 0.0001;

  float theta = 30;
  sint = sin(theta);
  cost = cos(theta);

  next();
}

void next() {
  buf.background(0, 0, 0);

  float r = 2.5;
  pa = random(-r, r);
  pb = random(-r, r);

  pc = random(-r, r);
  pd = random(-r, r);

}

void mousePressed() {
  loop();
  next();
}


void draw() {
  drawIntoBuffer();
  int avg = buf.getAverage();
  int target = avg*16;
  if(target > buf._clipTop>>4) noLoop();
  buf.render(target);
}

void drawIntoBuffer() {
  int s = ceil(sqrt(ncps));
  float is = 2.0/s;
  for(int i = ncps; i > 0; i--) {
    float x = 2*(i/s)/(float)s -1 + random(0,is);
    float y = 2*(i%s)/(float)s -1 + random(0,is);
    
    float p = zin(random(-PI,PI));

    for (int j = niters; j > 0; j--) {
      float t = j/(float)niters;

      float xp = sin(pa*x) + sin(pb*y) + p;
      float yp = sin(pc*x) + sin(pd*y) + p;
      x = xp;
      y = yp;

      int b = floor(t*myv);
      int r = floor((1-t)*myv);
      int g = floor(2*t*(1-t)*myv);
      
      buf.add(cx + xp*sc, cy + yp*sc, r,g,b , mya);
    }
  }
}

float zin (float angle) {
  int s = 8;
  if(angle < 0) {
    return round(s*sin(angle))/(float)s;
  } 
  else
    return sin(angle);
}

class FixedPointBuffer { // 8.16 fixed point

  int _w, _h, _area;
  int [] _rBuf;
  int [] _gBuf;
  int [] _bBuf;

  int _clipTop;

  FixedPointBuffer (int w, int h) {
    _w = w;
    _h = h;
    _area = w * h;
    _rBuf = new int[_area];
    _gBuf = new int[_area];
    _bBuf = new int[_area];
    _clipTop = (256<<16)-1;

    loadPixels();
  }

  synchronized void background(int r, int g, int b) {
    for(int idx = 0; idx < _area; idx++) {
      _rBuf[idx] = r;
    }
    for(int idx = 0; idx < _area; idx++) {
      _gBuf[idx] = g;
    }
    for(int idx = 0; idx < _area; idx++) {
      _bBuf[idx] = b;
    }
  }

  synchronized void add(float x, float y, int r, int g, int b, float a) {
    // convert to continuous coordinates
    x -= 0.5;
    y -= 0.5;

    // bounds check
    if((x < 0.0) || (y < 0.0) || (x >= _w - 1.0) || (y >= _h - 1.0)) return;

    // integral coordinates
    int ix1 = (int)x;
    int iy1 = (int)y;
    int ix2 = ix1 + 1;
    int iy2 = iy1 + 1;

    // fractional coordinates
    float fractx = x - (float)ix1;
    float fracty = y - (float)iy1;

    // reciprocal(?) of fractional coordinates
    float recipx = 1.0 - fractx;
    float recipy = 1.0 - fracty;

    // preconvert color values to floats
    float fr = (float)r;
    float fg = (float)g;
    float fb = (float)b;

    // plot it
    float ratio;
    int idx, c;


    // upper-left
    ratio = recipx * recipy * a;
    idx = iy1 * _w + ix1;
    c = (int)((ratio * fr) + _rBuf[idx]); 
    if(c>_clipTop) c=_clipTop; 
    _rBuf[idx] = c;
    c = (int)((ratio * fg) + _gBuf[idx]); 
    if(c>_clipTop) c=_clipTop; 
    _gBuf[idx] = c;
    c = (int)((ratio * fb) + _bBuf[idx]); 
    if(c>_clipTop) c=_clipTop; 
    _bBuf[idx] = c;

    // upper-right
    ratio = fractx * recipy * a;
    idx = iy1 * _w + ix2;
    c = (int)((ratio * fr) + _rBuf[idx]); 
    if(c>_clipTop) c=_clipTop; 
    _rBuf[idx] = c;
    c = (int)((ratio * fg) + _gBuf[idx]); 
    if(c>_clipTop) c=_clipTop; 
    _gBuf[idx] = c;
    c = (int)((ratio * fb) + _bBuf[idx]); 
    if(c>_clipTop) c=_clipTop; 
    _bBuf[idx] = c;

    // lower-left
    ratio = recipx * fracty * a;
    idx = iy2 * _w + ix1;
    c = (int)((ratio * fr) + _rBuf[idx]); 
    if(c>_clipTop) c=_clipTop; 
    _rBuf[idx] = c;
    c = (int)((ratio * fg) + _gBuf[idx]); 
    if(c>_clipTop) c=_clipTop; 
    _gBuf[idx] = c;
    c = (int)((ratio * fb) + _bBuf[idx]); 
    if(c>_clipTop) c=_clipTop; 
    _bBuf[idx] = c;

    // lower-right
    ratio = fractx * fracty * a;
    idx = iy2 * _w + ix2;
    c = (int)((ratio * fr) + _rBuf[idx]); 
    if(c>_clipTop) c=_clipTop; 
    _rBuf[idx] = c;
    c = (int)((ratio * fg) + _gBuf[idx]); 
    if(c>_clipTop) c=_clipTop; 
    _gBuf[idx] = c;
    c = (int)((ratio * fb) + _bBuf[idx]); 
    if(c>_clipTop) c=_clipTop; 
    _bBuf[idx] = c;
  }


  int getAverage() {
    float total = 0;

    for (int idx = _area -1; idx >=0; idx--) {
      total += _rBuf[idx]+_gBuf[idx]+_bBuf[idx];
    }

    return ceil(total/(_area*3));
  }

  void render(int target) {

    int up = max(1,_clipTop / target);

    for (int idx = _area -1; idx >=0; idx--) {
      pixels[idx] = 0xFF000000 |
        (min(_clipTop,_rBuf[idx]*up) & 0xFF0000) |
        ((min(_clipTop,_gBuf[idx]*up) & 0xFF0000) >> 8) |
        ((min(_clipTop,_bBuf[idx]*up) & 0xFF0000) >> 16);

    }
    updatePixels();
  }
}


























